(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(********************************************************************
*  builder prend en paramtre le scheduler et une liste de modules 
*********************************************************************)
let builder modules = 
  let sched =
    Aux.option_filter
      (function Mast.SCHED(s) -> Some s | _ -> None) modules in
  let modules =
    Aux.option_filter
      (function Mast.MODL(m) -> Some m | _ -> None) modules in
  match sched with
    [Mast.SCHED_MODULE(n,d,s,u,g,p,c,adm,h,f,ln)] ->
(*      (Order_crit.extract_oc c); *)
      let (handler_fv_env,interface_fv_env) = Variables.process modules in
      List.iter Timercheck.timer_check modules;
      let modules = Signature.verif_signature u s modules in
      let modules = Requires.global_requires_function g p modules in
      let criteria = Order_crit.verif_order_crit_param c modules in
      let admit = Madmit.process modules adm in
      let aspects =
	Aux.flat_map
	  (function
	      Mast.MODULE(nm,states,constants,types,glbls,field,extfuns,
			  criteria,admit,handlers,inter,fns,aspects,a) ->
			    aspects)
	  modules in
      let constants =
	Aux.flat_map
	  (function
	      Mast.MODULE(nm,states,constants,types,glbls,field,extfuns,
			  criteria,admit,handlers,inter,fns,aspects,a) ->
			    List.map Mast2ast.valdef constants)
	  modules in
      let types = Types.collect_types modules in
      let external_functions =
	Aux.union_list
	  (Aux.flat_map
	     (function
		 Mast.MODULE(nm,states,constants,types,glbls,field,extfuns,
			     criteria,admit,handlers,inter,fns,aspects,a) ->
			       List.map Mast2ast.extfun extfuns)
	     modules) in
      let globals =
	let tmp =
	  Aux.flat_map
	    (function
		Mast.MODULE(nm,states,constants,types,glbls,field,extfuns,
			    criteria,admit,handlers,inter,fns,aspects,a) ->
		  List.filter
		    (function
			Mast.VALDEF(Mast.VARDECL(Mast.NOT_REQUIRES,_,_,_,_),
				    _,ic,a)
		      | Mast.SYSDEF(Mast.VARDECL(Mast.NOT_REQUIRES,_,_,_,_),
				    ic,a) -> true
		      | _ -> false)
		    glbls)
	    modules in
	List.map Mast2ast.valdef tmp in
      let attributes =
	let tmp =
	  Aux.flat_map
	    (function
		Mast.MODULE(nm,states,constants,types,glbls,field,extfuns,
			    criteria,admit,handlers,inter,fns,aspects,a) ->
		  List.filter
		    (function
			Mast.VARDECL(Mast.NOT_REQUIRES,_,_,_,_) -> true
		      | _ -> false)
		    field) 
	    modules in
	List.map Mast2ast.vardecl tmp in
      Ast.SCHEDULER(Objects.id2c n,d,constants (* constants *),
		    List.map Mast2ast.typedef types,
		    attributes,external_functions,globals,
		    List.map Mast2ast.state s,
		    List.map Mast2ast.oc criteria,
		    Mast2ast.admit admit, Ast.NOTRACE,
		    List.map Mast2ast.event_decl
		      (Super.verif_sched_hd_construction h
			 handler_fv_env modules),
		    List.map Mast2ast.function_decl
		      (Super.verif_sched_fn_construction f
			 interface_fv_env modules),
		    [],
		    Ast.INITASPECTS(List.map Mast2ast.aspect aspects),
		    ln)
  | _ -> raise (Error.Error "exactly one scheduler should be provided")
