(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* For every unblock.timer.X handler, X has to be declared either as an
attribute or as a global variable *)

module B = Objects

let error mn str =
  Error.update_error();
  Printf.printf "%s: %s \n" (B.id2c mn) str

let pr = Printf.sprintf

let check_handlers mn valdefs procdecls handlers =
  let timer_handlers =
    Aux.option_filter
      (function
	  Mast.EVENT(Mast.EVENT_NAME([unblock;timer;timer_name],false,a),
		    specd,param,overrideable,stm,attr)
	  when B.id2c unblock = "unblock" && B.id2c timer = "timer" ->
	    Some timer_name
	| _ -> None)
      handlers in
  let timers_with_targets =
    Aux.option_filter
      (function
	  Mast.VARDECL(_,B.TIMER(_),id,_,_) -> Some id
	| _ -> None)
      procdecls in
  let timers_without_targets =
    Aux.option_filter
      (function
	  Mast.SYSDEF(Mast.VARDECL(_,B.TIMER(_),id,_,_),_,_) -> Some id
	| _ -> None)
      valdefs in
  let timers = Aux.union timers_with_targets timers_without_targets in
  List.iter
    (function timer ->
      if not (List.mem timer timer_handlers)
      then error mn (pr "no handler provided for timer %s" (B.id2c timer)))
    timers;
  List.iter
    (function timer_handler ->
      if not (List.mem timer_handler timers)
      then error mn (pr "no timer declaration provided for timer handler %s"
		       (B.id2c timer_handler)))
    timer_handlers
  
let timer_check
    (Mast.MODULE(mn,_,_,_,globals,attributes,_,_,_,handlers,_,_,_,_)) =
  check_handlers mn globals attributes handlers
