(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* Compute a checksum describing the parameters of all of the interface
functions.  Nothing protects this from overflow, but hopefully there are not
so many interface functions. *)

module B = Objects

let checksum_id id =
  let str = B.id2c id in
  let sum = ref 0 in
  String.iter (function c -> sum := Char.code c + (!sum)) str;
  !sum

let rec checksum_type = function
    B.ENUM(identifier) -> 1 + checksum_id identifier
  | B.RANGE(identifier) -> 2 + checksum_id identifier
  | B.SYSTEM(typ) -> 3 + checksum_type typ
  | B.INT -> 4
  | B.BOOL -> 5
  | B.TIME -> 6
  | B.CYCLES -> 7
  | B.PROCESS -> 8
  | B.SRCPROCESS -> 9
  | B.SCHEDULER -> 10
  | B.VOID -> 11
  | B.EVENT -> 12
  | B.INDR(typ) -> 13 + checksum_type typ
  | B.STRUCT(identifier) -> 14 + checksum_id identifier
  | B.NULL -> 15
  | B.EMPTY -> 16
  | B.TIMER(B.NOT_PERSISTENT) -> 17
  | B.TIMER(B.PERSISTENT(Some id)) -> 18 + checksum_id id
  | B.TIMER(B.PERSISTENT(None)) -> 19
  | B.PORT -> 20
  | B.SCHEDULER_INST -> 21

let checksum_decl (Ast.VARDECL(ty,id,_,_)) =
  checksum_type ty + checksum_id id
    
let checksum_fundef (Ast.FUNDEF(ty,nm,params,_,_,_)) index =
  checksum_type ty + (index * checksum_id nm) +
    (List.fold_left
       (function n -> function decl -> n + checksum_decl decl)
       0 params)

let checksum functions =
  let (n,index) =
    List.fold_left
      (function (n,index) ->
	function fundef -> (n + checksum_fundef fundef index,index+1))
      (0,1) functions in
  n
