(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: compile_interface.ml,v 1.12 2004/11/11 16:15:00 julia Exp $ *)

module B = Objects
module G = Genaux
module A = Ast2c
module CS = Class_state

open Format

(* ------------------------- basic functions ------------------------- *)

let make_start checksum =
  G.print_nl_string
    "// --------------------- interface functions ---------------------";
  G.print_nl_string ("// checksum = "^(string_of_int checksum))

let make_includes _ =
  G.print_nl_string "#include <linux/types.h>";
  G.print_nl_string "//#include <linux/errno.h>";
  G.print_nl_string "//#include <linux/sched.h>";
  G.print_nl_string "#include <linux/fs.h>";
  G.print_nl_string "#include <linux/unistd.h>";
  G.print_nl_string "//#include <linux/slab.h>";
  G.print_nl_string "#include <linux/devfs_fs_kernel.h>";
  G.print_nl_string "#include <asm/uaccess.h>";
  G.print_nl_string "#include <linux/ioctl.h>"

let make_uincludes nm =
  if (!B.csharp) then (
    G.print_nl_string "//needs the shared library libbossa.so obtains by : gcc -shared .../bossa/load.c -o /usr/lib/libbossa.so";
    G.print_nl_string "";
    G.print_nl_string "using System;";
    G.print_nl_string "using System.Runtime.InteropServices;";
    G.print_nl_string "using System.Threading;";
    G.print_nl_string "";
    G.print_nl_string ("public class user_"^nm);
    G.print_nl_string "{";
    G.print_nl_string "";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int bossa_attach_info(int parent);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int bossa_connect(int parent, int child, int argc, int *argv);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int bossa_connect_root(int new_root, int argc, int *argv);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int bossa_disconnect(int tgt);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int bossa_print(string s, int sz);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int scheduler_names(string s, int sz);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe void load_module(string s);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int find_scheduler(string b, string s);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe void mount(string parent_name, string child_name,";
    G.print_nl_string "\t\t\tint child_property_count,";
    G.print_nl_string "\t\t\tint *child_properties);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe void mount_root(string new_root_name,";
    G.print_nl_string "\t\t\tint child_property_count,";
    G.print_nl_string "\t\t\tint *child_properties);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe void unmount(string tgt_name);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int open(";
    G.print_nl_string "\t\tstring filename,";
    G.print_nl_string "\t\tuint flags);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int ioctl(int d,";
    G.print_nl_string "\t\tuint requete,";
    G.print_nl_string "\t\tvoid* lpInBuffer);";
    G.print_nl_string "\t\t//int lpCompletionRoutine);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe void close(int hFile);";
    G.print_nl_string "";
    G.print_nl_string "\t[DllImport(\"bossa\", SetLastError=true, CharSet=CharSet.Unicode)]";
    G.print_nl_string "\tstatic extern unsafe int system(string command);";
    G.print_nl_string "";
    G.print_nl_string "public static unsafe int install_module(string path) {";
    G.print_nl_string "\treturn system(\"cil2module \" + path);";
    G.print_nl_string "}";
    G.print_nl_string "";
    G.print_nl_string "\tconst uint GENERIC_READ  = 0x80000000;";
    G.print_nl_string "\tconst uint GENERIC_WRITE = 0x40000000;";
    G.print_nl_string "\tconst uint OPEN_EXISTING = 3;";
    G.print_nl_string "\tconst uint FILE_FLAG_NO_BUFFERING = 0x20000000;";
    G.print_nl_string "\t//const  int SIZE_INT = System.Runtime.InteropServices.Marshal.SizeOf(0);";
    G.print_nl_string "\tconst  int SIZE_INT = 4;";
    G.print_nl_string "\tconst uint IOCTL_BASE = (1U << 30) | (BOSSA_ID_NUM << 8) | (SIZE_INT << 16);";
  )else(
    G.print_nl_string "#include <stdio.h>";
    G.print_nl_string "#include <sys/ioctl.h>";
    G.print_nl_string "#include <sys/types.h>";
    G.print_nl_string "#include <sys/stat.h>";
    G.print_nl_string "#include <fcntl.h>";
    G.print_nl_string "#include <errno.h>";
    G.print_nl_string "#include <sched.h>";
    G.print_nl_string "#include <sys/time.h>";
    G.print_nl_string "#include <sys/resource.h>";
    G.print_nl_string ("#include \"user_stub_"^nm^".h\"");
  )

let choose_ctr nm ctr trace_ctr =
  match nm with
    "attach" -> 255
  | "start_trace" | "stop_trace" | "print_trace" ->
      let cur = !trace_ctr in
      trace_ctr := (!trace_ctr) + 1;
      255-cur
  | _ ->
      let cur = !ctr in
      ctr := (!ctr) + 1;
      cur

(* actually the number should not always be 0xCE *)
let make_ioctl_decls nm functions rw =
  if (!B.csharp) then G.print_nl_string "const uint BOSSA_ID_NUM = 0xCE;"
  else G.print_nl_string "#define BOSSA_ID_NUM 0xCE";
  let defstr =
    if (!B.csharp) then "static uint IOCTL_" else "#define IOCTL_" in
  let maxlen =
    (String.length defstr) +
      5 + (* leave some space *)
      List.fold_left
      (function maxlen ->
	function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) ->
	  let len = String.length (B.id2c nm) in
	  if len > maxlen then len else maxlen)
      0 functions in
  let ctr = ref 0 in
  let trace_ctr = ref 1 in
  List.iter2
    (function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) ->
      function (rd,wr) ->
	let snm = B.id2c nm in
	let str = String.make maxlen ' ' in
	let addstr = defstr^(String.uppercase snm) in
	let strlen = String.length addstr in
	let cur_ctr = choose_ctr snm ctr trace_ctr in
	String.blit addstr 0 str 0 strlen;
	print_string str;
	B.init_ioctls :=
	  ((!B.init_ioctls) ^ "    IOCTL_" ^ (String.uppercase(snm)));
	if (!B.csharp) then
	(G.print_nl_string ";";
	(match (rd,wr) with
  	    (rd,[]) -> B.init_ioctls := ((!B.init_ioctls) ^ "  =  __IOW(")
	  | ([],wr) -> B.init_ioctls := ((!B.init_ioctls) ^ "  =  __IOR(")
	  | _ -> B.init_ioctls := ((!B.init_ioctls) ^ "  =  __IOWR("));
	  B.init_ioctls := ((!B.init_ioctls) ^ "BOSSA_ID_NUM, ");
	  B.init_ioctls := ((!B.init_ioctls) ^ Int64.to_string(Int64.of_int(cur_ctr)) ^ ");\n")
	)
	else
	  ((match (rd,wr) with
  	    (rd,[]) -> print_string "_IOW("
	  | ([],wr) -> print_string "_IOR("
	  | _ -> print_string "_IOWR(");
	  print_string "BOSSA_ID_NUM, ";
	  print_int cur_ctr;
	  G.print_nl_string ", int)"))
    functions rw;
  if (!B.csharp)
  then G.print_nl_string ("private const string _BOSSA_DEVICE_NAME = \"bossa_"^nm^"\";")
  else G.print_nl_string ("#define BOSSA_DEVICE_NAME \"bossa_"^nm^"\"")


let make_ioctl_decls_cs nm functions rw =
  G.print_nl_string "\tconst uint BOSSA_ID_NUM = 0xCE;";
  let defstr = "const uint IOCTL_" in
  let maxlen =
    (String.length defstr) +
      5 + (* leave some space *)
      List.fold_left
      (function maxlen ->
	function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) ->
	  let len = String.length (B.id2c nm) in
	  if len > maxlen then len else maxlen)
      0 functions in
  let ctr = ref 0 in
  List.iter2
    (function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) ->
      function (rd,wr) ->
	let str = String.make maxlen ' ' in
	let addstr = defstr^(String.uppercase(B.id2c nm)) in
	let strlen = String.length addstr in
	String.blit addstr 0 str 0 strlen;
	print_string ("\t"^str^" = IOCTL_BASE | (");
	print_int (!ctr);
	G.print_nl_string " << 0);";
 	ctr := !ctr + 1;
    )
    functions rw;
  G.print_nl_string ("\tprivate const string _BOSSA_DEVICE_NAME = \"bossa_"^nm^"\";")



let make_find_process_by_pid name =
 if (!B.csharp)
 then (
   G.print_nl_string "[MethodImplAttribute(MethodImplOptions.InternalCall)]";
   G.print_nl_string
     ("private unsafe static extern "^(G.pp_bossa_struct_cs())^
      " * find_process_by_pid(int i);"))
 else (
   let process_struct =
     B.sched_select (G.pp_bossa_struct_cs()) "struct bossa_struct" in
   print_string
     ("static inline "^process_struct^
      " *find_process_by_pid(pid_t pid) ");
   G.start_block();
   G.print_nl_string "struct task_struct *tsk;";
   force_newline();
   print_string "if (pid) ";
   G.start_block();
   G.print_nl_string "//read_lock_irq(&tasklist_lock);";
   G.print_nl_string "tsk = find_task_by_pid(pid);";
   print_string "//read_unlock_irq(&tasklist_lock);";
   G.end_block();
   force_newline();
   G.print_nl_string "else tsk = current;";
   print_string ("return ("^process_struct^" *)(&tsk->bossa);");
   G.end_block())

let make_file_operations name =
  print_string "static struct file_operations procfops = ";
  G.start_block();
  print_string ("ioctl: dev_"^name^"_ioctl");
  G.end_block();
  G.pp_semic()
    
let make_major _ =
  G.print_nl_string "static int major;";
  if (!B.csharp) then (
    G.print_nl_string "static devfs_entry* devfs_handle;";
    G.print_nl_string "static devfs_entry* devfs_handle2;";
   ) else (
    G.print_nl_string "static devfs_handle_t devfs_handle;";
    G.print_nl_string "static devfs_handle_t devfs_handle2;";
    G.print_nl_string "extern spinlock_t bossa_scheduler_lock;"
   )
      
let make_ioctl_init name =
  let dev_name =
    if (!B.csharp) then "_BOSSA_DEVICE_NAME" else "BOSSA_DEVICE_NAME" in
  G.mk_static(); print_string "int ";
  if not (!B.csharp) then print_string "__init ";
  print_string (name^"_ioctl_init");
  G.mk_empty_prototype(); print_string " "; G.start_block();
  G.print_nl_string "devfs_handle = devfs_mk_dir (null, \"bossa\", null);";
  G.print_nl_string
    ("major = devfs_register_chrdev(0, "^dev_name^", &procfops);");
  G.print_nl_string "if (major < 0) {";
  print_string "  ";
  G.mk_printk "devfs_register_chrdev failed\\n";
  if (!B.csharp)
  then G.print_nl_string "  return -1;"
  else G.print_nl_string "  return -ENODEV;";
  G.print_nl_string "}";
  if (!B.csharp)
  then
    G.print_nl_string
      "devfs_handle2 = __devfs_register(devfs_handle, _BOSSA_DEVICE_NAME, major, 0, ref procfops);"
  else (
    G.print_nl_string
      "devfs_handle2 = devfs_register(devfs_handle, BOSSA_DEVICE_NAME, DEVFS_FL_DEFAULT,";
    G.print_nl_string "                                major, 0,";
    G.print_nl_string "                                S_IFCHR | S_IRUGO | S_IWUGO,";
    G.print_nl_string "                                &procfops, NULL);");
  print_string "return 0;";
  G.end_block()

(* ------------------------------ Errors ----------------------------- *)

let get_errno _ =
  if !B.csharp then "get_errno()" else "errno"

let fail_message str err_res =
  G.mk_printk str;
  if !B.csharp
  then G.print_nl_string "set_errno(-_EINVAL);"
  else G.print_nl_string "errno = -EINVAL;";
  print_string ("return "^err_res^";")

(* --------------------- analyze parameter usage --------------------- *)
      
let rec rw_exp params = function
    Ast.INT(n,attr) -> []
	
  | Ast.VAR(id,attr) ->
      if (Aux.member id params)
      then [id]
      else []
	  
  | Ast.FIELD(exp1,fld,attr) -> rw_exp params exp1
	
  | Ast.BOOL(v,attr) -> []
	
  | Ast.UNARY(uop,uexp,attr) -> rw_exp params uexp
	
  | Ast.BINARY(bop,exp1,exp2,attr) ->
      (rw_exp params exp1) @ (rw_exp params exp2)
			       
  | Ast.PBINARY(bop,exp1,states1,exp2,states2,attr) ->
      raise
	(Error.Error "internal error: compile_interface: pbinary unexpected")
			       
  | Ast.INDR(exp,attr) -> rw_exp params exp
	
  | Ast.SELECT(_,attr) -> []
	
  | Ast.EMPTY(id,_,_,_,attr) -> []
	
  | Ast.PRIM(f,args,attr) ->
      (rw_exp params f) @ (Aux.flat_map (rw_exp params) args)
			    
  | Ast.SCHEDCHILD(arg,procs,attr) -> []
	
  | Ast.SRCONSCHED(attr) -> []

  | Ast.ALIVE(_,_,attr) -> []
	
  | Ast.AREF(exp1,exp2,attr) ->
      (rw_exp params exp1) @ (rw_exp params exp2)
			       
  | Ast.IN(exp1,id,_,_,_,attr) -> rw_exp params exp1
	
  | Ast.MOVEFWDEXP(exp,state_end,attr) -> rw_exp params exp
	
let rec rw_stmt params = function
    Ast.IF(exp,stmt1,stmt2,attr) ->
      let (rd1,wr1) = rw_stmt params stmt1
      and (rd2,wr2) = rw_stmt params stmt2 in
      ((rw_exp params exp) @ rd1 @ rd2, wr1 @ wr2)
	
  | Ast.FOR(id,state_id,states,dir,stmt,attr) ->
      rw_stmt (Aux.remove id params) stmt
	
  | Ast.FOR_WRAPPER(label,stmts,attr) ->
      let (rd1, wr1) = List.split (List.map (rw_stmt params) stmts) in
      (List.fold_left Aux.union [] rd1, List.fold_left Aux.union [] wr1)
	
  | Ast.SWITCH(exp,cases,default,attr) ->
      let (rd,wr) = rw_cases params cases in
      let (rd1,wr1) =
	(match default with
	  None -> ([],[])
	| Some x -> rw_stmt params x) in
      ((rw_exp params exp) @ rd1 @ rd, wr1 @ wr)
	
  | (Ast.SEQ(decls,stmts,attr) as x) ->
      let (params, rd) =
	List.fold_left
	  (function (params,rd) ->
	    function
		(Ast.VALDEF(Ast.VARDECL(ty,id,imported,attr),exp,is_const,
			    attr1)) ->
			      let rd1 = rw_exp params exp in
			      let params = Aux.remove id params in
			      (params, rd1@rd)
	      | (Ast.UNINITDEF(Ast.VARDECL(ty,id,imported,attr),attr1))
		-> (Aux.remove id params, rd)
	      | (Ast.SYSDEF(Ast.VARDECL(ty,id,imported,attr),isconst,attr1))
		-> (Aux.remove id params, rd)
	      | (Ast.DUMMYDEF(id,exp,attr)) ->
		  let rd1 = rw_exp params exp in
		  let params = Aux.remove id params in
		  (params, rd1@rd))
	  (params, []) decls in
      let (rd1, wr1) = List.split (List.map (rw_stmt params) stmts) in
      (List.fold_left
	 (function res -> function rd -> rd @ res)
	 [] (rd::rd1),
       List.fold_left
	 (function res -> function wr -> wr @ res)
	 [] wr1)
	
  | Ast.RETURN(None,attr) -> ([],[])
	
  | Ast.RETURN(Some(exp),attr) -> (rw_exp params exp, [])
	
  | Ast.MOVE(exp,state,_,_,_,_,_,attr) -> (rw_exp params exp, [])
	
  | Ast.MOVEFWD(exp,_,_,attr) -> (rw_exp params exp, [])

  | Ast.SAFEMOVEFWD(exp,_,stm,state_end,attr) ->
      raise
	(Error.Error
	   "compile_interface: unexpected safe move forward statement\n")
	
  | Ast.ASSIGN(Ast.VAR(id,_) as expl,expr,sorted_fld,attr) ->
      ((rw_exp params expl) @ (rw_exp params expr),
       if Aux.member id params then [id] else [])
	
  | Ast.ASSIGN(expl,expr,sorted_fld,attr) ->
      ((rw_exp params expl) @ (rw_exp params expr), [])
	
  | Ast.PRIMSTMT(f,args,attr) ->
      ((rw_exp params f) @ (Aux.flat_map (rw_exp params) args), [])
	
  | Ast.ASSERT(exp,_) -> (rw_exp params exp , [])

  | x -> ([],[])
	
and rw_cases params cases =
  List.fold_left
    (function (rd,wr) ->
      function Ast.SEQ_CASE(idlist,_,stmt,_) ->
	let (crd,cwr) = rw_stmt params stmt in
	(crd@rd,cwr@wr))
    ([],[]) cases
    
let rw_function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) =
  let params =
    List.map (function (Ast.VARDECL(ty,id,imported,attr)) -> id) params in
  rw_stmt params stmt
    
(* -------- Data structures for interface function arguments --------- *)
    
let pp_interface_decl = function
    (Ast.VARDECL(B.PROCESS,id,imported,attr))
  | (Ast.VARDECL(B.SCHEDULER,id,imported,attr)) ->
      (if (!B.csharp) then print_string "public ";
       print_string ("int "^(B.id2c id)))
  | decl -> (if (!B.csharp) then print_string "public ";
	     A.pp_decl false decl)

let pp_interface_decl_2 = function
    (Ast.VARDECL(B.PROCESS,id,imported,attr))
  | (Ast.VARDECL(B.SCHEDULER,id,imported,attr)) ->
      print_string ("int "^(B.id2c id))
  | (Ast.VARDECL(B.TIME,id,imported,attr)) ->
      print_string
	("unsigned long "^(B.id2c id)^"_sec, long "^(B.id2c id)^"_nsec")
  | decl -> (A.pp_decl false decl)

let make_input_structure dev_name user
    (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) =
  let nm = B.id2c nm in
  if nm = "attach" && not user && !B.schedtype = B.VIRTSCHED
  then
    (let struct_name = nm^"_struct" in
    print_string ("struct "^dev_name^"_"^struct_name^" ");
    G.start_block();
    G.print_nl_string "int parent;";
    G.print_nl_string "int child;";
    print_string "int data[10];";
    G.end_block();
    G.pp_semic();
    force_newline())
  else
    (if (not (params = []))
    then
      if nm = "detach"
      then ()
      else
	(let struct_name = nm^"_struct" in
	print_string ("struct "^dev_name^"_"^struct_name^" ");
	G.start_block();
	if nm = "attach"
	then
	  begin
	    print_string "int _checksum";
	    if not (params = []) then G.pp_semic()
	  end;
	G.print_between G.pp_semic pp_interface_decl params;
	print_string ";";
	G.end_block();
	G.pp_semic();
	force_newline()))
      
(* ------------------------ User-declared types ----------------------- *)

(* don't print ENUMVALDEFS as these don't come from the source program *)
  
let pp_typedefs typedefs =
  let enum_function id f idlist =
    if (!B.csharp) then print_string "public ";
    print_string "enum"; print_space(); G.pp_id(id);
    print_space();
    if (!B.csharp) then print_string ": int ";
    G.start_block();
    G.print_between G.pp_comma f idlist;
    G.end_block();
    if (!B.csharp)
    then force_newline()
    else G.pp_semic() in
  List.iter
    (function
	Ast.ENUMDEF(id,idlist,attr) -> enum_function id G.pp_id idlist
      |	Ast.ENUMVALDEF(id,idlist,attr) -> ()
	    (*
	       enum_function id
	       (function (id,exp) ->
	         G.pp_id id; print_string " = "; A.pp_exp exp)
	       idlist
	     *)
      | Ast.RANGEDEF(id,exp1,exp2,attr) ->
	  (print_string "typedef int "; G.pp_id(id); G.pp_semic())
      | Ast.TIMERDEF(id,decls,attr) -> ()
	  (* these can't be used in stub functions *))
    typedefs; G.nl_ifnonempty(typedefs)
    
(* ----------------- Processing an interface function ---------------- *)
    
let generate_read_arguments locals params rd dev_name struct_name err_res
    stmtfn =
  if locals
  then
    List.iter
      (function (Ast.VARDECL(ty,id,imported,attr)) as decl ->
	if Aux.member id rd
	then
	  begin
	    A.pp_decl false decl;
	    G.pp_semic();
	    if (!B.schedtype = B.VIRTSCHED && ty = B.SCHEDULER)
	    then
	      G.print_nl_string
		("struct scheduler_operations *"^(B.id2c id)^"_tmp;")
	  end)
      params;
  print_string "copy_from_user";
  if (!B.csharp) then
    G.pp_arglist true
      ["(int*)&data"; "(int*)param";
       "sizeof("^dev_name^"_"^struct_name^")"]
  else
    G.pp_arglist true
      ["&data";
       "(struct "^dev_name^"_"^struct_name^" *)param";
       "sizeof(struct "^dev_name^"_"^struct_name^")"];
  G.pp_semic();
  let processes =
    List.fold_left
      (function processes ->
	function (Ast.VARDECL(ty,id,imported,attr)) as decl ->
	  if Aux.member id rd
	  then
	    (match ty with
	      B.PROCESS ->
		G.pp_id id;
		print_string " = ";
		B.check_proc "process paramater";
		print_string "find_process_by_pid(data.";
		G.pp_id id;
		G.print_nl_string ");";
		(id,id)::processes
	    | B.SCHEDULER ->
		let id_tmp = B.mkId((B.id2c id)^"_tmp") in
		G.pp_id id_tmp;
		print_string " = ";
		B.check_virt "scheduler paramater";
		print_string "find_scheduler_by_id(data.";
		G.pp_id id;
		G.print_nl_string ");";
		(id,id_tmp)::processes
	    | _ ->
		G.pp_id id;
		print_string " = ";
		print_string "data.";
		G.pp_id id;
		G.pp_semic();
		processes)
	  else processes)
      [] params in
  let ty = B.sched_select "process" "scheduler" in
  List.iter
    (function (process,process_tmp) ->
      print_string ("if ("^(B.id2c process_tmp)^"==null) ");
      G.start_block();
      fail_message ("unknown "^ty^"\\n")
	err_res;
      G.end_block();
      if locals (* not in attach *)
      then
	(force_newline();
	 let sprc = B.id2c process in
	 let sprc_tmp = B.id2c process_tmp in
	 (match !B.schedtype with
	   B.PROCSCHED ->
	     print_string
	       ("if ("^sprc_tmp^"->next_list != &(EXPORTS.next_list)) ")
	 | B.VIRTSCHED ->
	     G.print_nl_string
	       "// the following should be ok because the scheduler is not the root";
	     print_string
	       ("if ("^sprc_tmp^
		"->next_list.prev->prev->sched->ops != &EXPORTS) "));
	 G.start_block();
	 fail_message
	   (ty^" argument "^sprc^
	    " must be managed by the current scheduler\\n")
	   "FAILURE";
	 G.end_block();
	 if !B.schedtype = B.VIRTSCHED
	 then
	   begin
	     let s_ok = sprc ^ "_ok" in
	     let stmp2 = sprc ^ "_tmp2" in
	     force_newline();
	     A.pp_for (B.mkId stmp2)
	       (function _ ->
		 print_string ("if ("^stmp2^"->ops == "^sprc_tmp^") ");
		 G.start_block();
		 G.print_nl_string (sprc^" = "^stmp2^";");
		 print_string ("goto "^s_ok^";");
		 G.end_block())
	       (function _ -> raise (Error.Error "internal error (for)"))
	       false
	       (List.map (function (_,s) -> s) (!CS.state_env))
	       Ast.RANDOM;
	     fail_message "scheduler not found\\n" "FAILURE";
	     force_newline();
	     print_string (s_ok^":")
	   end);
      force_newline())
    processes;
  stmtfn()
      
let pp_attach_ioctl checksum policy_name = function
    (Ast.VARDECL(ty,id,imported,attr)) :: _ ->
      G.start_block();
      G.print_nl_string
	(G.de_structify (G.pp_bossa_struct_cs()^" before_attach;"));
      G.print_nl_string
	(G.de_structify (G.pp_bossa_struct_cs()^" after_attach;"));
      force_newline();
      G.print_nl_string "error_string = \"\";";
      if (!B.csharp)
      then G.print_nl_string "if (get_errno() != 0) "
      else print_string "if (errno != 0) ";
      G.start_block();
      fail_message "some error has previously occurred\\n" "FAILURE";
      G.end_block();
      force_newline();
      G.print_nl_string
	"// 0 checksum allowed for eg hourglass, that doesn't use attach";
      print_string
	("if (data._checksum && data._checksum != "^
	 (string_of_int checksum)^") ");
      G.start_block();
      fail_message
	("stub functions are out of date, attach to "^policy_name^
	 " refused\\n")
	"FAILURE";
      G.end_block();
      force_newline();
      print_string
	("if ("^(B.id2c id)^"->next_list == &(EXPORTS.next_list)) ");
      G.start_block();
      fail_message
	("can't move from a scheduler ("^policy_name^") to itself\\n")
	"FAILURE";
      G.end_block();
      force_newline();
      G.print_nl_string "set_current_need_resched(1);";
      if (!B.csharp) then (
	G.print_nl_string
	  ("memcpy((int*)&before_attach, (int*)"^(B.id2c id)^
	   ",sizeof("^G.pp_bossa_struct_cs()^"));");
        G.print_nl_string
	  ("EXPORTS.next_list.next->sched->ops->ps_attach("^
	   "EXPORTS.next_list.next->next);");
        G.print_nl_string
	  ("memcpy((int*)&after_attach, (int*)"^(B.id2c id)^
	   ",sizeof("^G.pp_bossa_struct_cs()^"));");
        G.print_nl_string
	  ("memcpy((int*)"^(B.id2c id)^
	   ", (int*)&before_attach,sizeof("^G.pp_bossa_struct_cs()^"));");
        print_string "if (get_errno() == 0) ";
	G.start_block();
        G.print_nl_string "set_root_s_state((SchedState)";
        print_string
  	  ("  "^(B.id2c id)^"->next_list->next->sched->ops->ps_detach");
        G.pp_arglist true [(B.id2c id)^"->next_list->next->next";
			   "(bossa_struct *)"^(B.id2c id)];
        print_string ")";
        G.pp_semic();
        G.print_nl_string
 	  ("memcpy((int*)"^(B.id2c id)^
	   ", (int*)&after_attach,sizeof("^G.pp_bossa_struct_cs()^"));");)
      else (
        G.print_nl_string
	  ("memcpy(&before_attach,"^(B.id2c id)^
	   ",sizeof("^G.pp_bossa_struct_cs()^"));");
        G.print_nl_string
	  ("EXPORTS.next_list.next->sched->ops->ps_attach("^
	   "EXPORTS.next_list.next->next);");
        G.print_nl_string
	  ("memcpy(&after_attach,"^(B.id2c id)^
	   ",sizeof("^G.pp_bossa_struct_cs()^"));");
        G.print_nl_string
	  ("memcpy("^(B.id2c id)^
	   ",&before_attach,sizeof("^G.pp_bossa_struct_cs()^"));");
        print_string "if (errno == 0) ";
	
        G.start_block();
        G.print_nl_string "root_s_state =";
        print_string
  	  ("  "^(B.id2c id)^"->next_list->next->sched->ops->ps_detach");
        G.pp_arglist true [(B.id2c id)^"->next_list->next->next";
			   "(struct bossa_struct *)"^(B.id2c id)];
        G.pp_semic();
        G.print_nl_string
 	  ("memcpy("^(B.id2c id)^
	   ",&after_attach,sizeof("^G.pp_bossa_struct_cs()^"));"));

      G.print_nl_string ((B.id2c id)^"->next_list = &(EXPORTS.next_list);");
      print_string ("if ("^get_errno()^" != 0) ");
      G.start_block();
      fail_message "Some error has occurred in detach\\nThe original scheduler may now be in an inconsistent state\\n" "FAILURE";
      G.end_block();
      G.end_block();
      force_newline();
      print_string "else ";
      G.start_block();
      G.print_nl_string "printk(\"%s\",error_string);";
      fail_message ("Some error has occurred in attach to "^policy_name^"\\n")
	"FAILURE";
      G.end_block();
      G.end_block();
      force_newline();
      print_string "return SUCCESS;"
  | _ -> raise (Error.Error "internal error 1")
	
let num_function = ref 0
    
let compile_function policy_name dev_name checksum
    (Ast.FUNDEF(ret,nm,params,stmt,inl,attr), (rd,wr)) =
  let nm = B.id2c nm in
  let struct_name = nm^"_struct" in
  let in_attach = nm = "attach" in
  let state_update_needed = not in_attach && A.move_possible stmt in
  let label = "out_"^nm in
  let phase = !A.current_phase in
  if state_update_needed || not (wr = [])
  then A.current_phase := A.INTERFACE_LABEL(label);
  if (!B.csharp)
  then
    (print_string "case "; print_int !num_function; print_string ": ";
     num_function := !num_function +1)
  else print_string ("case IOCTL_"^(String.uppercase nm)^": ");
  G.start_block();
  if (not (params = [])) then
    G.print_nl_string
      (G.de_structify ("struct "^dev_name^"_"^struct_name^" data;"));
  (* decode the arguments *)
  if not(rd = [])
  then
    generate_read_arguments (not in_attach) params rd dev_name struct_name
      "FAILURE"
      (if in_attach
      then (function () -> pp_attach_ioctl checksum policy_name params)
      else (function () -> A.pp_stmt stmt))
  else A.pp_seq stmt;
  A.current_phase := phase;
  if state_update_needed || not (wr = [])
  then
    begin
      force_newline();
      print_string label; G.print_nl_string ":";
      if state_update_needed
      then
	(if (!B.csharp)
	 then G.print_nl_string "set_root_s_state((SchedState)"
	 else G.print_nl_string "root_s_state =";
	 print_string "  EXPORTS.next_list.next->sched->ops->compute_state(EXPORTS.next_list.next->next)";
	 if (!B.csharp) then print_string ")";
	 G.pp_semic());
      if not (wr = [])
      then
	(List.iter
	   (function x ->
	     G.print_nl_string
	       (Printf.sprintf "data.%s = %s;" (B.id2c x) (B.id2c x)))
	   wr;
	 print_string "copy_to_user";
	 if (!B.csharp) then
	   G.pp_arglist true
	     ["(int*)param"; "(int*)&data";
	       "sizeof("^dev_name^"_"^struct_name^")"]
	 else
	   G.pp_arglist true
	     ["(struct "^dev_name^"_"^struct_name^" *)param";
	       "&data";
	       "sizeof(struct "^dev_name^"_"^struct_name^")"];
	 G.pp_semic());
(*
   G.print_nl_string
   "if (root_s_state != RUNNING) current->need_resched = 1;"*)
      print_string "return SUCCESS;"
    end;
  G.end_block()
    
let compile_ps_attach dev_name (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) =
  let struct_name = "attach_struct" in
  G.mk_static(); print_string "int ";
  print_string(dev_name^"_ps_attach");
  G.pp_arglist false ["struct next_list_struct *next"];
  print_string " ";
  let any_time_parameters =
    List.filter
      (function (Ast.VARDECL(B.TIME,_,_,_)) -> true
	| _ -> false)
      params in
  if any_time_parameters = []
  then A.pp_seq stmt
  else
    begin
      G.start_block();
      List.iter
	(function (Ast.VARDECL(_,id,_,_)) ->
	  let id = B.id2c id in
	  G.print_nl_string
	    (id^".tv_nsec = tstojiffie("^id^".tv_sec,"
	     ^id^".tv_nsec,"^id^".tv_sec);"))
	any_time_parameters;
      A.pp_seq stmt;
      G.end_block()
    end;
  force_newline();
  force_newline()
    
let compile_vs_attach name (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) =
  let nm = B.id2c nm in
  let struct_name = nm^"_struct" in
  let (child_param,params) =
    match params with
      (Ast.VARDECL(ty,id,imported,attr)) :: other_params -> (id,other_params)
    | _ -> raise (Error.Error "internal error 2") in
  G.print_nl_string "static int";
  print_string(name^"_attach");
  G.pp_arglist true [(G.pp_bossa_struct_cs())^" *"^(B.id2c child_param);
		     "int data[]"];
  print_string " ";
  G.start_block();
  G.print_nl_string "int _i = 0;";
  List.iter
    (function (Ast.VARDECL(ty,id,imported,attr)) as decl ->
      A.pp_decl false decl;
      print_string " = ";
      G.print_nl_string "data[_i++];")
    params;
  A.pp_stmt stmt;
  G.end_block();
  force_newline();
  force_newline();
  print_string("static int "^name^"_attach_info(char *info[]) ");
  G.start_block();
  let _ =
    List.fold_left
      (function n ->
	function Ast.VARDECL(ty,id,_,_) ->
	  G.print_nl_string
	    (Printf.sprintf "info[%d] = \"%s %s\";"
	       n (B.type2c ty) (B.id2c id));
	  n+1)
      0 params in
  print_string (Printf.sprintf "return %d;" (List.length params));
  G.end_block();
  force_newline();
  force_newline()
    
let compile_detach dev_name (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) =
  match params with
    [Ast.VARDECL(B.PROCESS,id,imported,attr)] ->
      G.mk_static();
      print_string ("int "^dev_name^"_ps_detach");
      G.pp_arglist true ["struct next_list_struct *next";
			 "struct bossa_struct *_"^(B.id2c id)];
      print_string " ";
      G.start_block();
      G.print_nl_string
  	((G.pp_bossa_struct_cs())^" *"^(B.id2c id)^
	 " = ("^(G.pp_bossa_struct_cs())^" *)_"^(B.id2c id)^";");
      A.pp_seq stmt;
      G.end_block();
      force_newline();
      force_newline()
  | [Ast.VARDECL(B.SCHEDULER,id,imported,attr);
     Ast.VARDECL(B.BOOL,id1,imported1,attr1)] ->
       G.mk_static(); print_string ("int "^dev_name^"_detach");
       G.pp_arglist true [(G.pp_bossa_struct_cs())^" *_"^(B.id2c id);
			  "int "^(B.id2c id1)];
       print_string " ";
       G.start_block();
       if (!B.csharp) then
         G.print_nl_string
	   ((G.pp_bossa_struct_cs())^" *"^(B.id2c id)^
	    " = ("^(G.pp_bossa_struct_cs())^" *)_"^(B.id2c id)^";")
       else
	 G.print_nl_string
	   ((G.pp_bossa_struct_cs())^" *"^(B.id2c id)^
	    " = ("^(G.pp_bossa_struct_cs())^" *)_"^(B.id2c id)^";");
       A.pp_seq stmt;
       G.end_block();
       force_newline();
       force_newline()
  | _ -> raise (Error.Error "internal error 3")
	
	
let compile_ifunctions nm functions rw checksum =
  (* declare attach variables *)
  if !B.schedtype = B.PROCSCHED
  then
    List.iter
      (function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) ->
	if (B.id2c nm) = "attach"
	then
	  List.iter (function x -> A.pp_decl true x; G.pp_semic()) params)
      functions;
  force_newline();
  let print_ioctl_params _ =
    open_box 0;
    G.print_between
      (function _ -> G.print_nl_string ",")
      print_string
      ["uint ioctl_num";"ulong param"];
    close_box() in
  G.mk_static(); G.mk_inline(); print_string ("int _dev_"^nm^"_ioctl");
    G.pp_arglist true ["uint ioctl_num";"ulong param"];
    print_string " ";
    G.start_block();
  if (!B.csharp) then (
    G.print_nl_string "set_errno(0);";
    (* check that the scheduler has been loaded *)
    print_string "if (EXPORTS.next_list.next == &(EXPORTS.next_list)) ";
    G.start_block();
    fail_message (nm^" not loaded\\n") "FAILURE";
    G.end_block();
    force_newline();
    print_string "switch (ioctl_num-";
    print_string (String.sub (!B.init_ioctls)
    		    ((String.index (!B.init_ioctls) '=') + 3)
		    ((String.index (!B.init_ioctls) ';')-((String.index (!B.init_ioctls) '=') + 3))
		 );
    
    print_string ") ";
   ) else (
    G.print_nl_string "errno = 0;";
    (* check that the scheduler has been loaded *)
    print_string "if (EXPORTS.next_list.next == &(EXPORTS.next_list)) ";
    G.start_block();
    fail_message (nm^" not loaded\\n") "FAILURE";
    G.end_block();
    force_newline();
    print_string "switch (ioctl_num) ";
   );
  G.start_block();
  G.print_between
    force_newline
    (if (!B.csharp)
    then compile_function nm ("___dev_"^nm) checksum
    else compile_function nm ("dev_"^nm) checksum)
    (List.combine functions rw);
  G.end_block();
  force_newline();
  fail_message ("error occurred in "^nm^"\\n") "FAILURE";
  G.end_block();
  force_newline();
  force_newline();
  G.mk_static(); print_string ("int dev_"^nm^"_ioctl");
    G.pp_arglist true ["struct inode *inode";"struct file *fd";
		       "uint ioctl_num";"ulong param"];
    print_string " ";
    G.start_block();
    G.print_nl_string "int res;";
    G.print_nl_string "ulong flags=0;";
    force_newline();
  if (!B.csharp) then (
    G.print_nl_string "spin_lock_irqsave(get_bossa_scheduler_lock(),flags);";
    G.print_nl_string
      ("res = _dev_"^nm^"_ioctl(ioctl_num,param);");
    G.print_nl_string "spin_unlock_irqrestore(get_bossa_scheduler_lock(),flags);";
    print_string ("return res;");
    G.end_block()
   ) else (
    G.print_nl_string "spin_lock_irqsave(&bossa_scheduler_lock,flags);";
    G.print_nl_string
      ("res = _dev_"^nm^"_ioctl(ioctl_num,param);");
    G.print_nl_string "spin_unlock_irqrestore(&bossa_scheduler_lock,flags);";
    print_string ("return res;");
    G.end_block()
   )
      
(* ------------------ Interface functions for users ----------------- *)
      
let ucompile_function policy_name dev_name checksum
    (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) =
  let nm = B.id2c nm in
  if not(nm = "detach")
  then
    (let struct_name = nm^"_struct" in
    if (!B.csharp) then print_string "public static unsafe ";
    print_string ("int "^policy_name^"_"^nm);
    print_string "(";
    open_box 0;
    if params = []
    then print_string "void"
    else G.print_between G.pp_comma pp_interface_decl_2 params;
    close_box();
    print_string ") ";
    G.start_block();
    (if params = [] && not(nm = "attach")
    then G.print_nl_string "int _data;"
    else
      if (!B.csharp)
      then G.print_nl_string (dev_name^"_"^struct_name^" _data;")
      else G.print_nl_string ("struct "^dev_name^"_"^struct_name^" _data;"));
    if nm = "attach"
    then G.print_nl_string ("_data._checksum = "^(string_of_int checksum)^";");
    List.iter
      (function
	  (Ast.VARDECL(B.TIME,id,imported,attr)) ->
	    print_string "_data."; G.pp_id id; print_string ".tv_sec = ";
	    G.pp_id id; print_string "_sec"; G.pp_semic();
	    print_string "_data."; G.pp_id id; print_string ".tv_nsec = ";
	    G.pp_id id; print_string "_nsec"; G.pp_semic()
	| (Ast.VARDECL(ty,id,imported,attr)) as decl ->
	print_string "_data."; G.pp_id id; print_string " = "; G.pp_id id;
	G.pp_semic())
      params;
    G.start_block();
    G.print_nl_string "int fd, res;";
    if (!B.csharp) then (
      G.print_nl_string
	"fd = open(\"/dev/bossa/\" + _BOSSA_DEVICE_NAME, GENERIC_READ | GENERIC_WRITE);";
      G.print_nl_string
	("res = ioctl(fd,IOCTL_"^(String.uppercase nm)^",&_data);");
     ) else (
      G.print_nl_string
	"fd = open(\"/dev/bossa/\" BOSSA_DEVICE_NAME, O_WRONLY);";
      G.print_nl_string
	("res = ioctl(fd,IOCTL_"^(String.uppercase nm)^",(char *)&_data);");
     );
    
    let tst_str =
      if ret = B.VOID then if (!B.csharp) then "!=0" else "" else " < 0" in
    let error_func =
      if (!B.csharp) then "Console.Error.WriteLine" else "perror" in
    G.print_nl_string ("if (res"^tst_str^") "^error_func^"(\""^nm^"\");");
    G.print_nl_string "close(fd);";
    print_string "return res;";
    G.end_block();
    G.end_block())
      
let hcompile_function dev_name (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) =
  let nm = B.id2c nm in
  if not ((!B.schedtype = B.VIRTSCHED && nm = "attach") or nm = "detach")
  then
    (print_string ("int "^dev_name^"_"^nm);
     print_string "(";
     open_box 0;
     if params = []
     then print_string "void"
     else G.print_between G.pp_comma pp_interface_decl_2 params;
     close_box();
     G.print_nl_string ");")
      
(* --------------------------- Entry point -------------------------- *)
      
let classify_ifunctions ifunctions =
  let (attach,detach,ifunctions,sfunctions (* needs a structure *)) =
    List.fold_left
      (function (attach,detach,ifunctions,sfunctions) ->
	function (Ast.FUNDEF(ret,nm,params,stmt,inl,attr)) as fd ->
	  match (B.id2c nm,!B.schedtype) with
	    ("attach",B.PROCSCHED) ->
	      (Some fd,detach,fd::ifunctions,fd::sfunctions)
	  | ("attach",B.VIRTSCHED) ->
	      (Some fd,detach,ifunctions,fd::sfunctions)
	  | ("detach",_) -> (attach,Some fd,ifunctions,sfunctions)
	  | _ -> (attach,detach,fd::ifunctions,fd::sfunctions))
      (None,None,[],[]) ifunctions in
  (attach,detach,List.rev ifunctions,List.rev sfunctions)
    
let compile_interface
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   _,trace,handlers,ifunctions,functions,_,attr))
    checksum =
  let (attach,detach,ifunctions,sfunctions) = classify_ifunctions ifunctions in
  A.current_phase := A.INTERFACE;
  make_start checksum;
  force_newline();
  force_newline();
  make_major();
  force_newline();
  force_newline();
  let rw = List.map rw_function ifunctions in
  make_ioctl_decls nm ifunctions rw;
  force_newline();
  force_newline();
  make_find_process_by_pid nm;
  force_newline();
  force_newline();
  if (!B.csharp)
  then List.iter (make_input_structure ("___dev_"^nm) false) sfunctions
  else List.iter (make_input_structure ("dev_"^nm) false) sfunctions;
  compile_ifunctions nm ifunctions rw checksum;
  force_newline();
  force_newline();
  if not (!B.csharp) then make_file_operations nm;
  force_newline();
  force_newline();
  make_ioctl_init nm;
  force_newline();
  force_newline();
  (match (attach,!B.schedtype) with
    (None,_) -> ()
  | (Some x,B.PROCSCHED) ->
      A.current_phase := A.ATTACH;
      compile_ps_attach nm x
  | (Some x,B.VIRTSCHED) ->
      A.current_phase := A.INTERFACE;
      compile_vs_attach nm x);
  (match detach with
    None -> ()
  | Some x ->
      if !B.schedtype = B.PROCSCHED
      then A.current_phase := A.DETACH;
      compile_detach nm x);
  A.current_phase := A.INTERFACE
      
let interface_includes _ =
  G.print_nl_string "/* includes for interface functions */";
  make_includes();
  force_newline();
  force_newline()
    
let ucompile_interface
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   _,trace,handlers,ifunctions,functions,_,attr))
    checksum =
  let (attach,detach,ifunctions,sfunctions) = classify_ifunctions ifunctions in
  A.current_phase := A.INTERFACE;
  make_start checksum;
  force_newline();
  force_newline();
  make_uincludes nm;
  force_newline();
  force_newline();
  let rw = List.map rw_function ifunctions in
  if (!B.csharp) then (
    make_ioctl_decls_cs nm ifunctions rw;
    force_newline();
    force_newline();
    pp_typedefs enums;
   ) else
    make_ioctl_decls nm ifunctions rw;
  force_newline();
  force_newline();
  List.iter (make_input_structure ("dev_"^nm) true) sfunctions;
  G.print_between (function _ -> force_newline(); force_newline())
    (ucompile_function nm ("dev_"^nm) checksum) ifunctions;
  force_newline();
  if (!B.csharp) then (G.print_nl_string "}"; force_newline();)
      
let hcompile_interface
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   _,trace,handlers,ifunctions,functions,_,attr)) =
  A.current_phase := A.INTERFACE;
  let policy_flag = "BOSSA_"^(String.uppercase nm) in
  G.print_nl_string ("#ifndef "^policy_flag);
  G.print_nl_string ("#define "^policy_flag);
  G.print_nl_string "#ifndef BOSSA_TYPEDEFS";
  G.print_nl_string "#define BOSSA_TYPEDEFS";
  G.print_nl_string "typedef int bool;";
  G.print_nl_string "typedef void * port;";
  G.print_nl_string "#endif";
  force_newline();
  pp_typedefs enums;
  force_newline();
  force_newline();
  List.iter (hcompile_function nm) ifunctions;
  force_newline();
  G.print_nl_string "#endif"
