%{
(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* $Id: bossa_parser.mly,v 1.15 2004/11/11 16:15:01 julia Exp $ *)

module B = Objects

%}

%token  SCHEDULER TYPE ENUM PROCESS CONST ORDERING LOWEST HIGHEST KEY TRACE
%token  STATES READY RUNNING BLOCKED POLICY_BLOCKED KERNEL_BLOCKED
%token  KERNEL_POLICY_BLOCKED TERMINATED QUEUE FIFO LIFO 
%token  HANDLER EVENT ON INTERFACE FUNCTION DEFAULT HIGHRES LOWRES
%token  INT BOOL TIME CYCLES VOID IF ELSE FOREACH IN SWITCH RETURN TIMER PORT
%token  PERSISTENT
%token  HEAD TAIL FOREACHINC FOREACHDEC
%token  CASE TRUE FALSE AND OR DOT LPAREN RPAREN LCURLY RCURLY BREAK
%token  DOT_DOT LBRACK RBRACK
%token  EQUAL_EQUAL NEQ COMMA SEMICOLON COLON EQL PLUS
%token  MINUS TIMES DIV GT LT COMPLEMENT LEQ GEQ EQUAL
%token  RARROW  BITOR BITAND PLUS_PLUS MINUS_MINUS SELECT EMPTY SRCONSCHED
%token  SCHEDULEROF
%token  PLUS_EQUAL MINUS_EQUAL TIMES_EQUAL DIV_EQUAL MOD_EQUAL ALIVE
%token  BITAND_EQUAL BITOR_EQUAL LSHIFT_EQUAL RSHIFT_EQUAL MOD NOT
%token  LSHIFT RSHIFT ERROR
%token  EVENTS TEST EXPRESSIONS
%token  FWD VIRTSCHEDULER DEFER PUBLIC PRIVATE QUESTION SYSTEM STRUCT
%token  ADMIT ADMIT_CRIT ADMIT_ATT ADMIT_DET
%token EOF
%token <int> INTEGER
%token <string> IDENT
%token <string> ENUM_TYPE
%token <string> RANGE_TYPE
%token <string> TIMER_TYPE
%token <string> FN_NAME
%token <string> STATE_NAME
%token <string> STRING

%start bossa_spec
%type <Ast.scheduler * Objects.sched> bossa_spec

/* specify operator precedence to allow simple single production rule for expressions */
%left COMMA                 /* lowest precedence */
%left EQL PLUS_EQUAL MINUS_EQUAL TIMES_EQUAL DIV_EQUAL MOD_EQUAL 
BITAND_EQUAL BITOR_EQUAL LSHIFT_EQUAL RSHIFT_EQUAL RARROW
%left OR
%left AND 
%left BITOR
%left BITAND
%left EQUAL_EQUAL NEQ
%left GT LT GEQ LEQ
%left LSHIFT RSHIFT
%left PLUS MINUS
%left TIMES DIV MOD IN
%right NOT COMPLEMENT 
%left PLUS_PLUS MINUS_MINUS
%left DOT
%nonassoc UMINUS                /* highest precedence */

/* start syntax definition */
%%
bossa_spec: 
    x_setpos is_default_hr SCHEDULER x_setproc IDENT EQL
    LCURLY scheddef RCURLY EOF
    { let 
	((c,t,p,v,s,o,adm,tr),(h,i,f)) = $8
      in (Ast.mkSCHEDULER($5,$2,c,t,p,v,s,o,adm,tr,h,i,f,Misc.getpos 10),
	  B.PROCSCHED)
    } 
|   x_setpos is_default_hr VIRTSCHEDULER x_setvirt IDENT EQL LCURLY vscheddef
    RCURLY EOF  
    { let 
	((c,t,p,v,s,o,adm,tr),(h,i,f)) = $8
      in (Ast.mkSCHEDULER($5,$2,c,t,p,v,s,o,adm,tr,h,i,f,Misc.getpos 10),
	  B.VIRTSCHED)
    } 
    ;

/* ----------------------------------------------------------------- */
/* high_res/low_res and default in any order */

is_default_hr:
    DEFAULT is_hr                       { true }
  | HIGHRES is_default
      { if (not(!Objects.high_res))
      then
	raise
	  (Error.Error
	     "high-res scheduler must be compiled with the option -hrt");
      $2}
  | LOWRES is_default
      { if (!Objects.high_res)
      then
	raise 
	  (Error.Error
	     "low-res scheduler may not be compiled with the option -hrt");
	$2}
  |                                     { false }
  ;

is_hr:
    HIGHRES
      { if (not(!Objects.high_res))
      then
	raise
	  (Error.Error
	     "high-res scheduler must be compiled with the option -hrt") }
  | LOWRES
    { if (!Objects.high_res)
    then
      raise 
	(Error.Error
	   "low-res scheduler may not be compiled with the option -hrt") }
  |                                     { () }
  ;

is_default:
    DEFAULT                             { true }
  |                                     { false }
  ;

scheddef:
    sched_decl sched_body 		{($1, $2)}
     ;

vscheddef:
    vsched_decl sched_body 		{($1, $2)}
     ;

sched_decl:
    constdef typedef_decl processdef globaldef statedef orderdef admission
      tracedef
    { ($1, $2, $3, $4, $5, $6, $7, $8)}
    ;

vsched_decl:
    constdef typedef_decl schedtypedef globaldef statedef orderdef admission
      tracedef
    { ($1, $2, $3, $4, $5, $6, $7, $8)}
    ;

sched_body:
    handlerdef interfacedef functiondef	{($1,$2,$3)}
    ;

typedef_decl:
    enum_def typedef_decl              { $1::$2 }
  | range_def typedef_decl             { $1::$2 }
  | timer_def typedef_decl             { $1::$2 }
  |                                    { [] }
  ;

enum_def:
    x_setpos TYPE IDENT EQL ENUM LCURLY enum_body RCURLY SEMICOLON 
    {Progtypes.add_enum_type $3;
      Ast.mkENUMDEF($3, $7, Misc.getpos 9)}

enum_body:
   IDENT  				{ [$1] }
   | IDENT COMMA enum_body        	{ $1::$3 }
   ;

range_def:
    x_setpos TYPE IDENT EQL LBRACK expr DOT_DOT expr RBRACK SEMICOLON
    {Progtypes.add_range_type $3;
      Ast.mkRANGEDEF($3, $6, $8, Misc.getpos 9)}

timer_def:
    x_setpos IDENT TIMER EQL LCURLY timer_def_body RCURLY
    {Progtypes.add_timer_type $2;
      Ast.mkTIMERDEF($2, $6, Misc.getpos 7)}

timer_def_body:
    non_timer_process_var_decl SEMICOLON                { [$1] } 
  | non_timer_process_var_decl SEMICOLON timer_def_body   { $1::$3 }
  ;

/* no nested timers */
non_timer_process_var_decl:
  x_setpos type_expr IDENT
      { Ast.mkVARDECL($2,$3, Misc.getpos 3) }
  | x_setpos type_expr SYSTEM IDENT
      { Ast.mkSYSVARDECL($2,$4, Misc.getpos 4) }
   ;

processdef:
    PROCESS EQL LCURLY process_body RCURLY   { $4 }
   |                                         { [] }
  ;

schedtypedef:
    SCHEDULER EQL LCURLY process_body RCURLY { $4 }
   |                                         { [] }
  ;

process_body:
  process_var_decl SEMICOLON                { [$1] } 
  | process_var_decl SEMICOLON process_body { $1::$3 }
  ;

process_var_decl:
  x_setpos type_expr IDENT
      { Ast.mkVARDECL($2,$3, Misc.getpos 3) }
  | x_setpos TIMER IDENT
      { Ast.mkVARDECL(B.TIMER(B.NOT_PERSISTENT),$3, Misc.getpos 3) }
  | x_setpos PERSISTENT TIMER IDENT
      { Ast.mkVARDECL(B.TIMER(B.PERSISTENT(None)),$4, Misc.getpos 4) }
  | x_setpos PERSISTENT TIMER_TYPE TIMER IDENT
      { Ast.mkVARDECL(B.TIMER(B.PERSISTENT(Some(B.mkId $3))),$5,
		      Misc.getpos 5) }
  | x_setpos type_expr SYSTEM IDENT
      { Ast.mkSYSVARDECL($2,$4, Misc.getpos 4) }
   ;

param_var_decl:
  x_setpos type_expr IDENT
      { Ast.mkVARDECL($2,$3,Misc.getpos 3) }
   ;

const_var_decl:
  x_setpos bossa_type_expr IDENT
      { Ast.mkVARDECL($2,$3, Misc.getpos 3) }
   ;

constdef:
   x_setpos CONST const_var_decl EQL expr SEMICOLON constdef
        { Ast.mkCONSTDEF($3,$5,Misc.getpos 7) :: $7 }
  |     { [] }
   ;

parameter_types:
  param_type_list        { $1 }
  |                      { [] }
  ;

/* for function prototypes */
param_type_list:
  one_param_type                         { [$1] }
  | one_param_type COMMA param_type_list { $1::$3 }

one_param_type:
  type_expr                          { $1 }
  | TIMER                            { B.TIMER(B.NOT_PERSISTENT) }
  | PERSISTENT TIMER                 { B.TIMER(B.PERSISTENT(None)) }
  | PERSISTENT TIMER_TYPE TIMER      { B.TIMER(B.PERSISTENT(Some(B.mkId $2))) }

globaldef:
  x_setpos non_proc_type_expr IDENT EQL expr SEMICOLON globaldef
      {let ln = Misc.getpos 7 in
      Ast.VL(Ast.mkVALDEF(Ast.mkVARDECL($2,$3,ln),$5,ln)) :: $7 }
  | x_setpos non_proc_type_expr SYSTEM IDENT SEMICOLON globaldef
      {let ln = Misc.getpos 6 in
      Ast.VL(Ast.mkSYSDEF(Ast.mkSYSVARDECL($2,$4,ln),ln)) :: $6}
  | x_setpos non_proc_type_expr IDENT LPAREN parameter_types RPAREN SEMICOLON
      globaldef
      { Progtypes.add_fn_name $3;
	Ast.FN(Ast.mkFUNDECL($2,$3,$5,Misc.getpos 8)) :: $8 }
  | x_setpos VOID IDENT LPAREN parameter_types RPAREN SEMICOLON
      globaldef
      { Progtypes.add_fn_name $3;
	Ast.FN(Ast.mkFUNDECL(B.VOID,$3,$5,Misc.getpos 8)) :: $8 }
  | x_setpos TIMER IDENT SEMICOLON globaldef
        { let loc = Misc.getpos 5 in
	  Ast.VL(Ast.mkSYSDEF(Ast.mkVARDECL(B.TIMER(B.NOT_PERSISTENT),$3, loc),
			      loc))
	  :: $5 }
  |   { [] }
  ;

valdef:
  x_setpos type_expr IDENT EQL expr SEMICOLON valdef
	{let ln = Misc.getpos 7 in
	Ast.mkVALDEF(Ast.mkVARDECL($2,$3,ln),$5,ln) :: $7 }
  |     { [] }
   ;

orderdef:
   ORDERING EQL LCURLY order_body RCURLY { $4 }
  |                                      { [] }
   ;

order_body:
  | crit_decl   		  { [$1] }
  | crit_decl COMMA order_body    { $1::$3 }
  ;

crit_decl:
   x_setpos key crit_order IDENT     { Ast.mkCRIT_ID($2,$3,$4,Misc.getpos 4) }
  | x_setpos key crit_order LPAREN expr QUESTION expr COLON expr RPAREN
      { Ast.mkCRIT_TEST($2,$3,$5,$7,$9,Misc.getpos 10) }
  ;

key:
   KEY             { Ast.KEY(0,0) }
  |                { Ast.NOKEY }

crit_order:
   LOWEST          { Ast.LOWEST }
   | HIGHEST       { Ast.HIGHEST }


/* ------------------------------------------------------------------ */
/* admissibility criteria for modules */

admission:
  x_setpos ADMIT 
    EQL LCURLY valdef adm_crit adm_attach_detach RCURLY
    { let (crit_param,crit) = $6 in
    let (attach,detach) = $7 in
    Some(Ast.mkADMIT($5,crit_param,crit,attach,detach,Misc.getpos 8)) }  
  | { None }

adm_crit:
  ADMIT_CRIT parameters EQL LCURLY expr RCURLY         { ($2,$5) }

adm_attach_detach:
  ADMIT_ATT proc_param EQL seq_stmt ADMIT_DET proc_param EQL seq_stmt
      { (Some($2,$4),Some($6,$8)) }
  |   { (None,None) }

proc_param:
  x_setpos LPAREN proc_type IDENT RPAREN
      { Ast.mkVARDECL($3,$4,Misc.getpos 5) }

/* ------------------------------------------------------------------ */
/* trace */

tracedef:
   x_setpos TRACE INTEGER LCURLY trace_events trace_expressions trace_test
            RCURLY
       { Progtypes.add_fn_name "print_trace_info";
	 match $5@$6@$7 with
	 [] -> Ast.NOTRACE
       | all -> Ast.mkTRACE($3, all, Misc.getpos 8)}
   |   { Ast.NOTRACE }

trace_events:
   EVENTS EQL LCURLY event_types RCURLY SEMICOLON
       {[Ast.TRACEEVENTS($4)]}
   | {[]}

trace_expressions:
   x_setpos EXPRESSIONS EQL LCURLY fields RCURLY SEMICOLON
       {[Ast.TRACEEXPS(Ast.mkTRACEEXP($5,Misc.getpos 7))]}
   | {[]}

trace_test:
   TEST EQL LCURLY expr RCURLY SEMICOLON
       {[Ast.TRACETEST($4)]}
   | {[]}

fields:
   IDENT                       { [$1] }
   | IDENT COMMA fields        { $1 :: $3 }
   | STATE_NAME                { [$1] }
   | STATE_NAME COMMA fields   { $1 :: $3 }

statedef:
   STATES EQL LCURLY state_body RCURLY{ $4 }
   ;

state_body:
   state_decl 		    { [$1] }
   | state_decl state_body  { $1::$2 }
   ;

state_decl:
     x_setpos class_name IDENT COLON PROCESS SEMICOLON
      { B.check_proc "process";
	Progtypes.add_state_name $3;
	Ast.mkPROCESS($2,$3,Class_state.PUBLIC,Misc.getpos 6) }
   | x_setpos class_name IDENT COLON queue_visibility SCHEDULER SEMICOLON
      { B.check_virt "scheduler";
	Progtypes.add_state_name $3;
	Ast.mkPROCESS($2,$3,$5,Misc.getpos 7) }
   | x_setpos class_name IDENT COLON queue_visibility queue_type QUEUE SEMICOLON
      { Progtypes.add_state_name $3;
	Ast.mkQUEUE($2,$6,$3,$5,Misc.getpos 8) }
   | x_setpos class_name IDENT SEMICOLON
      { Progtypes.add_state_name $3;
	Ast.mkQUEUE($2,Class_state.EMPTY,$3,Class_state.PRIVATE,Misc.getpos 4)}
   ;

class_name:
   READY			{Class_state.READY}
   | RUNNING			{Class_state.RUNNING}
   | BLOCKED			{Objects.is_using_blocked();
				  match (!Objects.policy_blocked_types,
					 !Objects.schedtype) with
				    (true,B.PROCSCHED) ->
				      Class_state.KERNEL_BLOCKED
				  | _ ->
				      Class_state.BLOCKED}
   | POLICY_BLOCKED             {Objects.check_proc "POLICY_BLOCKED";
				  Objects.is_using_kernel_blocked();
				 Class_state.POLICY_BLOCKED}
   | KERNEL_BLOCKED             {Objects.check_proc "KERNEL_BLOCKED";
				  Objects.is_using_kernel_blocked();
				 Class_state.KERNEL_BLOCKED}
   | KERNEL_POLICY_BLOCKED	{Objects.check_proc "KERNEL_POLICY_BLOCKED";
				  Objects.is_using_kernel_blocked();
				 Class_state.KERNEL_POLICY_BLOCKED}
   | TERMINATED			{Class_state.TERMINATED}
   ;

queue_visibility:
   PUBLIC			{Class_state.PUBLIC}
   | PRIVATE			{Class_state.PRIVATE}
   |				{Class_state.PUBLIC}
   ;

queue_type:
  SELECT			{Class_state.SELECT(Class_state.NODEFAULT) }
  | SELECT FIFO			{Class_state.SELECT(Class_state.FIFODEFAULT) }
  | SELECT LIFO			{Class_state.SELECT(Class_state.LIFODEFAULT) }
  |                             {Class_state.FIFO }
  ;

handlerdef:
   HANDLER LPAREN EVENT IDENT RPAREN LCURLY event_body RCURLY
	{ List.map
	    (function x -> (x (Ast.mkVARDECL(B.EVENT,$4,-1))))
	    ($7) }
   ;
event_body:
   event_decl           	{ $1 }
  | event_decl event_body       { $1@$2 }
  ;

event_decl:
  x_setpos ON event_types seq_stmt
		{ let ln = Misc.getpos 4 in
		  List.map
		  (function nm ->
		    function param -> Ast.mkEVENT(nm,param,$4,ln))
		  $3 }
  ;

event_types:
  x_setpos event_type_   { [$2(Misc.getpos 2)]}
  | x_setpos event_type_ COMMA event_types   { ($2(Misc.getpos 4))::$4}
;

event_type_:
  event_name hevt_list          { function x ->
                                  let (l,s) = $2 in
                                  Ast.mkEVENT_NAME($1::l,s,x) }
  ;

event_name:
  IDENT                         { $1 }
  | STATE_NAME                  { $1 }
  | PROCESS                     { "process" }
  | SYSTEM                      { "system" }
  | TIMER                       { "timer" }




hevt_list:
  DOT TIMES			{ ([],true) }
  | DOT event_name hevt_list    { let (l,s) = $3 in ($2::l,s) }
  |                             { ([],false) }
  ;


interfacedef:
  INTERFACE EQL LCURLY interface_body RCURLY { $4 }
  ;

interface_body:
     intdef	                { [$1] }
   | intdef interface_body      { $1::$2 }
   ;

intdef:
  x_setpos type_or_void IDENT parameters seq_stmt
	{ Ast.mkFUNDEF($2,$3,$4,$5,Misc.getpos 5) }
  ;

functiondef:
  FUNCTION EQL LCURLY function_body RCURLY  { $4 }
  |                                         { [] }
  ;

function_body:
     fundef		       { [$1] }
   | fundef function_body      { $1::$2 }
   ;

fundef:
  x_setpos type_or_void FN_NAME parameters seq_stmt
	{ Ast.mkFUNDEF($2,$3,$4,$5,Misc.getpos 5) }
  ;

parameters:
 LPAREN param_lists_body RPAREN   { $2 }
 | LPAREN RPAREN 	          { [] }
 ;

param_lists_body:
   param_var_decl 	                   { [$1] }
   | param_var_decl COMMA param_lists_body { $1::$3 }
   ;

bossa_type_expr:
  base_type	 {$1}
  | proc_type	 {$1}
  | ENUM_TYPE 	 {B.ENUM(B.mkId($1))}
  | RANGE_TYPE 	 {B.RANGE(B.mkId($1))}
  ;

type_expr:
  bossa_type_expr         { $1 }
  | SYSTEM STRUCT IDENT   { B.SYSTEM(B.STRUCT(B.mkId($3))) }
  ;

non_proc_type_expr:
  base_type	          { $1 }
  | ENUM_TYPE             { B.ENUM(B.mkId($1)) }
  | RANGE_TYPE 	          { B.RANGE(B.mkId($1)) }
  | SYSTEM STRUCT IDENT   { B.SYSTEM(B.STRUCT(B.mkId($3))) }
  ;

base_type:
  INT				{ B.INT }
  | BOOL			{ B.BOOL }
  | TIME 			{ B.TIME }
  | CYCLES 			{ B.CYCLES }
  | PORT			{ B.PORT }
  ;

proc_type:
  | PROCESS			{ B.check_proc "process"; B.PROCESS }
  | SCHEDULER			{ B.check_virt "scheduler"; B.SCHEDULER }
  ;

type_or_void:
  type_expr		{ $1 }
  | VOID		{B.VOID}
  ;

stmt:
  if_stmt    		{ $1 }
  | for_stmt		{ $1 }
  | return_stmt 	{ $1 }
  | switch_stmt		{ $1 }
  | seq_stmt   		{ $1 }
  | assign_stmt	SEMICOLON	{ $1 }
  | defer_stmt          { $1 }
  | prim_stmt           { $1 }
  | error_stmt          { $1 }
  | break_stmt          { $1 }
  ;

if_stmt:
  x_setpos if_condition seq_stmt  
      { Ast.mkIF($2,$3, Ast.mkSEQ([],[],-1),Misc.getpos 3) }
  | x_setpos if_condition seq_stmt ELSE seq_stmt	
	{ Ast.mkIF($2,$3,$5,Misc.getpos 5) }
  
;

if_condition:
  IF LPAREN expr RPAREN  { $3 }
  ;

/* foreach increasing and decreasing is only supported when only one state
name (not even a class name that is inly implemented by one state) is given
*/
for_stmt:
  x_setpos FOREACH LPAREN IDENT IN state_name_list RPAREN seq_stmt 
	{Ast.mkFOR($4,$6,$8,Misc.getpos 8)}
  | x_setpos FOREACHINC LPAREN IDENT IN STATE_NAME RPAREN seq_stmt 
	{Ast.mkFORINC($4,$6,$8,Misc.getpos 8)}
  | x_setpos FOREACHDEC LPAREN IDENT IN STATE_NAME RPAREN seq_stmt 
	{Ast.mkFORDEC($4,$6,$8,Misc.getpos 8)}
  | x_setpos FOREACH LPAREN IDENT RPAREN seq_stmt 
	{Ast.mkFOR_ANY($4,$6,Misc.getpos 6)}
  ;

state_name_list :
   STATE_NAME  				{ [$1] }
   | class_name				{ [Class_state.class2idc $1] }
   | STATE_NAME COMMA state_name_list  	{ $1::$3 }
   | class_name COMMA state_name_list  	{ (Class_state.class2idc $1)::$3 }
   ;

switch_stmt:
  x_setpos SWITCH loc_expr IN seq_case { Ast.mkSWITCH($3,$5,Misc.getpos 5)}
  ;

return_stmt:
  x_setpos RETURN expr SEMICOLON		{ Ast.mkRETURN($3,Misc.getpos 4) }
  | x_setpos RETURN SEMICOLON		{ Ast.mkRETURNVOID(Misc.getpos 3) }
  ;

error_stmt:
  x_setpos ERROR LPAREN STRING RPAREN SEMICOLON
      {Ast.mkERROR($4,Misc.getpos 6)}

break_stmt:
  x_setpos BREAK SEMICOLON
      {Ast.mkBREAK(Misc.getpos 3)}

seq_case:
  LCURLY case_stmt RCURLY	{ $2 }
  ;

case_stmt:
  x_setpos CASE state_lists COLON seq_stmt case_stmt
      { Ast.mkSEQ_CASE($3,$5,Misc.getpos 6) :: $6 }
  |   { [] }
  ;

state_lists:
  STATE_NAME		         {[$1]}
  | STATE_NAME COMMA state_lists {$1::$3}
  | class_name		         {[Class_state.class2c($1)]}
  | class_name COMMA state_lists {(Class_state.class2c($1))::$3}
  ;

field_ref:
  IDENT	        {$1}
  ;

seq_stmt:
   x_setpos LCURLY valdef s_body RCURLY	{Ast.mkSEQ($3,$4,Misc.getpos 5)}
   ;

s_body:
  stmt s_body   {$1::$2}
  |		{[]}
  ;

defer_stmt:
   x_setpos DEFER LPAREN RPAREN SEMICOLON
      {B.check_virt "defer()"; Ast.mkDEFER(Misc.getpos 5) }

prim_stmt:
   x_setpos FN_NAME LPAREN arglist RPAREN SEMICOLON
      {Ast.mkPRIMSTMT($2,$4,Misc.getpos 6)}

assign_stmt:
  x_setpos assign_stmt_         { $2(Misc.getpos 2) }

assign_stmt_:
  loc_expr PLUS_PLUS
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.PLUS,Ast.mkINT(1,x),x)}
  | loc_expr MINUS_MINUS
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.MINUS,Ast.mkINT(1,x),x)}
  | loc_expr EQL expr {function x ->  Ast.mkASSIGN($1,$3,x)}
  | loc_expr PLUS_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.PLUS,$3,x)}
  | loc_expr MINUS_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.MINUS,$3,x)}
  | loc_expr TIMES_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.TIMES,$3,x)}
  | loc_expr DIV_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.DIV,$3,x)}
  | loc_expr MOD_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.MOD,$3,x)}
  | loc_expr BITAND_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.BITAND,$3,x)}
  | loc_expr BITOR_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.BITOR,$3,x)}
  | loc_expr LSHIFT_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.LSHIFT,$3,x)}
  | loc_expr RSHIFT_EQUAL expr
      {function x -> Ast.mkCOMPOUND_ASSIGN($1,Ast.RSHIFT,$3,x)}
/* This ought to allow an arbitrary expr to the left of the arrow, but
doing so causes parsing problems.  Fortunately it seems that only a loc_expr
or select() can be type correct. */
  | move_expr RARROW STATE_NAME
      {function x -> Ast.mkMOVE($1,$3,x)}
  | move_expr RARROW STATE_NAME DOT HEAD
      {function x -> Ast.mkMOVEHD($1,$3,x)}
  | move_expr RARROW STATE_NAME DOT TAIL
      {function x -> Ast.mkMOVETL($1,$3,x)}
  | move_expr RARROW FWD LPAREN RPAREN
      {B.check_virt "forwardImmediate()";
	function x -> Ast.mkMOVEFWD($1,x)}
  | move_expr RARROW FWD LPAREN RPAREN DOT HEAD
      {B.check_virt "forwardImmediate()";
	function x -> Ast.mkMOVEHDFWD($1,x)}
  | move_expr RARROW FWD LPAREN RPAREN DOT TAIL
      {B.check_virt "forwardImmediate()";
	function x -> Ast.mkMOVETLFWD($1,x)}
  ;

expr:
  x_setpos expr_                { $2(Misc.getpos 2) }

expr_:   
   INTEGER      {function x -> Ast.mkINT($1, x)}
   | IDENT      {function x -> Ast.mkVAR($1, x)}
   | STATE_NAME {function x -> Ast.mkVAR($1, x)}
   | TRUE	{function x -> Ast.mkBOOL(true, x)}
   | FALSE	{function x -> Ast.mkBOOL(false, x)}
   | PLUS expr_  {function x -> Ast.mkBINARY(Ast.PLUS, Ast.mkINT(0,x), $2 x,
			      x)}
   | MINUS expr_ {function x -> Ast.mkBINARY(Ast.MINUS, Ast.mkINT(0,x), $2 x,
			      x)}
   | TIMES expr_ {function x -> Ast.mkINDR($2 x, x)}
   | NOT expr_ {function x -> Ast.mkUNARY(Ast.NOT, $2 x, x)}
   | COMPLEMENT expr_  {function x -> Ast.mkUNARY(Ast.COMPLEMENT, $2 x, x)}
   | expr_ DOT field_ref  {function x -> Ast.mkFIELD($1 x,$3,x)}
   | SELECT LPAREN RPAREN {function x -> Ast.mkSELECT(x)}
   | FN_NAME LPAREN arglist RPAREN {function x -> Ast.mkPRIM($1,$3,x)}
   | EMPTY LPAREN STATE_NAME RPAREN {function x -> Ast.mkEMPTY($3,x)}
   | EMPTY LPAREN class_name RPAREN {function x -> Ast.mkEMPTYCLASS($3,x)}
   | SCHEDULEROF LPAREN expr RPAREN {function x -> Ast.mkSCHEDCHILD($3,x)}
   | ALIVE LPAREN RPAREN {function x -> Ast.mkALIVE(x)}
   | SRCONSCHED LPAREN RPAREN {function x -> Ast.mkSRCONSCHED(x)}
   | expr_ PLUS expr_ {function x -> Ast.mkBINARY(Ast.PLUS, $1 x, $3 x, x)}
   | expr_ MINUS expr_ {function x -> Ast.mkBINARY(Ast.MINUS, $1 x, $3 x, x)}
   | expr_ TIMES expr_ {function x -> Ast.mkBINARY(Ast.TIMES, $1 x, $3 x, x)}
   | expr_ DIV expr_ {function x -> Ast.mkBINARY(Ast.DIV, $1 x, $3 x, x)}
   | expr_ MOD expr_ {function x -> Ast.mkBINARY(Ast.MOD, $1 x, $3 x, x)}
   | expr_ AND expr_ {function x -> Ast.mkBINARY(Ast.AND, $1 x, $3 x, x)}
   | expr_ OR expr_ {function x -> Ast.mkBINARY(Ast.OR, $1 x, $3 x, x)}
   | expr_ BITAND expr_ {function x -> Ast.mkBINARY(Ast.BITAND, $1 x, $3 x, x)}
   | expr_ BITOR expr_ {function x -> Ast.mkBINARY(Ast.BITOR,$1 x, $3 x, x)}
   | expr_ EQUAL_EQUAL expr_ {function x -> Ast.mkBINARY(Ast.EQ,$1 x, $3 x, x)}
   | expr_ NEQ expr_ {function x -> Ast.mkBINARY(Ast.NEQ,$1 x, $3 x, x)}
   | expr_ LT expr_ {function x -> Ast.mkBINARY(Ast.LT, $1 x, $3 x, x)}
   | expr_ GT expr_ {function x -> Ast.mkBINARY(Ast.GT, $1 x, $3 x, x)}
   | expr_ LEQ expr_ {function x -> Ast.mkBINARY(Ast.LEQ,$1 x, $3 x, x)}
   | expr_ GEQ expr_ {function x -> Ast.mkBINARY(Ast.GEQ,$1 x, $3 x, x)}
   | expr_ LSHIFT expr_ {function x -> Ast.mkBINARY(Ast.LSHIFT, $1 x, $3 x, x)}
   | expr_ RSHIFT expr_ {function x -> Ast.mkBINARY(Ast.RSHIFT, $1 x, $3 x, x)}
   | expr_ IN STATE_NAME {function x -> Ast.mkIN($1 x, $3, x)}
   | expr_ IN class_name {function x -> Ast.mkINCLASS($1 x, $3, x)}
   | LPAREN expr RPAREN {function x -> $2}
    ;

arglist:
   arglist1                      { $1 }
   |                             { [] }
   ;

arglist1:
   expr                          { [$1] }
   | expr COMMA arglist1         { $1 :: $3 }

loc_expr:
   x_setpos loc_expr_             { $2(Misc.getpos 2) }

loc_expr_:
  IDENT         	{function x -> Ast.mkVAR($1, x)}
  | STATE_NAME   	{function x -> Ast.mkVAR($1, x)}
  | loc_expr_ DOT IDENT {function x -> Ast.mkFIELD($1 x, $3, x)}
  ;

move_expr:
  x_setpos SELECT LPAREN RPAREN      { Ast.mkSELECT(Misc.getpos 4) }
  | x_setpos SCHEDULEROF LPAREN expr RPAREN
      { Ast.mkSCHEDCHILD($4,Misc.getpos 5) }
  | loc_expr                        { $1 }

/* helper production. Purpose: sideeffect: set lineno to get position
where expr starts */
x_setpos:
	  { 
	    Stack.push (!Misc.lineno, !Misc.cr_index) Misc.pos_stack;
	  }  
	  ;
x_setvirt:
	  { 
	    B.schedtype := B.VIRTSCHED
	  }  
	  ;

x_setproc:
	  { 
	    B.schedtype := B.PROCSCHED
	  }  
	  ;
