(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* ------------- Process scheduler or virtual scheduler ------------- *)

type sched = PROCSCHED | VIRTSCHED
type sched_module = MODULE | PROCSCHED2
val schedtype : sched ref
val schedmodule : sched_module ref

val check_module : string -> unit

val check_proc : string -> unit
val check_virt : string -> unit
val sched_select : 'a -> 'a -> 'a

val check_blocked : string -> unit
val check_policy_blocked : string -> unit

(* --------------- Choice of blocked or kernel_blocked -------------- *)

type ps_using_blocked = B_UNKNOWN | IS_USING_BLOCKED | IS_USING_KERNEL_BLOCKED
val is_using_blocked : unit -> unit
val is_using_kernel_blocked : unit -> unit
val ps_using_blocked : ps_using_blocked ref

(* --------------------------- Identifiers -------------------------- *)

(* string for printing, int for comparison *)
type identifier = ID of string * int
val init_id : unit -> unit
val mkId : string -> identifier
val ideq : (identifier * identifier) -> bool
val strideq : (string * identifier) -> bool
val id2c : identifier -> string
val id2n : identifier -> int
val nopid : identifier
val fresh_id : unit -> identifier
val fresh_idx : string -> identifier

val symtab : ((string, identifier) Hashtbl.t) ref
val index : int ref
val fresh_ctr : int ref

(* ------------------------------ Types ----------------------------- *)

type typ =
  ENUM of identifier | RANGE of identifier | SYSTEM of typ |
  INT | BOOL | TIME | CYCLES | PROCESS | SRCPROCESS | SCHEDULER | VOID |
  EVENT | INDR of typ | STRUCT of identifier | NULL | EMPTY | PORT |
  SCHEDULER_INST | TIMER of timer_persistence

and timer_persistence = NOT_PERSISTENT | PERSISTENT of identifier option

val type2c : typ -> string     (* for debugging *)
val type2ctype : typ -> string (* for C code *)

(* ---------------------------- Attributes -------------------------- *)

(* test_index is used to correlate the occurrences of tests in specialized
handlers, to know when the test is true or false in all instances, to know
when to give an error message *)

type location = MODULE_NAME of identifier | SCHEDULER_NAME of identifier
type attribute = {modl : location ref option; ln : int; test_index : int;
		   ty : typ; com : string option}
val module_name : location ref option ref
val mkAttr : int -> attribute
val mkTestAttr : int * int -> attribute
val modl : attribute -> location ref option
val line : attribute -> int
val test_index : attribute -> int
val ty : attribute -> typ
val com : attribute -> string option
val updty : attribute -> typ -> attribute
val updtst : attribute -> int -> attribute
val add_comment : attribute -> string -> attribute
val loc2c : attribute -> string

(* ----------------- Optimizing empty and in tests ------------------ *)

type testval = TRUE | FALSE | BOTH | UNK
val lubtestval : testval * testval -> testval
val testval2c : testval -> string

(* -------------------------- For events ---------------------------- *)

type intinfo = INTIBLE | UNINTIBLE

val mutate : bool ref

val csharp : bool ref
val hls : bool ref
val blast : bool ref
val high_res : bool ref
val policy_blocked_types : bool ref

val init_ioctls : string ref

val type_return : string ref

type isquasi = QUASI | NOT_QUASI   (* detect quasi event types *)

(* -------------------------- Primitives ---------------------------- *)

val prim_names : unit -> string list

(* ------------------------ Free variables ------------------------- *)

val capture_check : (identifier * attribute) list ->
    identifier list -> (string -> unit) -> unit
