(*/////////////////////////////////////////////////////////////////////////////
//
// Bossa, a framework for scheduler programming
// Copyright (C) 2005 Ecole des Mines de Nantes and
// Department of Computer Science, University of Copenhagen
//
// This file is part of Bossa.
//
// Bossa is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// any later version.
//
// Bossa is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with Bossa; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Please send remarks, questions and bug reports to bossa@emn.fr
// or to:
//
//   Gilles Muller
//   Ecole des Mines de Nantes,
//   La Chantrerie - 4, rue Alfred Kastler. B.P. 20722
//   44307 NANTES Cedex 3, FRANCE
//
///////////////////////////////////////////////////////////////////////////////
*)

(* drop unnecessary nested sequences *)

let rec process_stm = function
    Ast.IF(tst,thn,els,a) ->
      [Ast.IF(tst,process_one_stm thn,process_one_stm els,a)]
  | Ast.FOR(x,y,z,d,stm,a) ->
      [Ast.FOR(x,y,z,d,process_one_stm stm,a)]
  | Ast.FOR_WRAPPER(x,stms,a) ->
      [Ast.FOR_WRAPPER(x,Aux.flat_map process_stm stms,a)]
  | Ast.SWITCH(exp,cases,dflt,a) ->
      let cases =
	List.map
	  (function Ast.SEQ_CASE(a,b,stm,c) ->
	    Ast.SEQ_CASE(a,b,process_one_stm stm,c))
	  cases in
      let dflt =
	match dflt with
	  None -> None
	| Some stm -> Some (process_one_stm stm) in
      [Ast.SWITCH(exp,cases,dflt,a)]
  | Ast.SEQ([],stms,a) -> Aux.flat_map process_stm stms
  | Ast.SEQ(decls,stms,a) ->
      [Ast.SEQ(decls,Aux.flat_map process_stm stms,a)]
  | Ast.SAFEMOVEFWD(exp,x,s,b,a) ->
      [Ast.SAFEMOVEFWD(exp,x,process_one_stm s,b,a)]
  | stm -> [stm]

and process_one_stm stm =
  match process_stm stm with 
    [x] -> x
  | stms -> Ast.SEQ([],stms,Ast.get_stm_attr stm)

(* --------------------------- Entry point -------------------------- *)

let process_handlers handlers =
  List.map
    (function Ast.EVENT(nm,specd,param,stm,attr) ->
      Ast.EVENT(nm,specd,param,process_one_stm stm,attr))
    handlers

let process_functions functions =
  List.map
    (function Ast.FUNDEF(ret,nm,params,stmt,inl,attr) ->
      Ast.FUNDEF(ret,nm,params,process_one_stm stmt,inl,attr))
    functions

let process_aspects = function
    Ast.ASPECTINFO(aspects) ->
      let process l =
	List.map
	  (function
	    Ast.ASPECT(srcs,src_infos,dsts,dst_infos,param,pos,stm,fvs,a) ->
	      Ast.ASPECT(srcs,src_infos,dsts,dst_infos,param,pos,
			 process_one_stm stm,fvs,a))
	  l in
      Ast.ASPECTINFO(
      List.map
	(function (transition,(bef,aft)) ->
	  (transition,(process bef,process aft)))
	aspects)
  | _ -> failwith "braces: unexpected aspects"

let process_admit = function
    None -> None
  | Some(Ast.ADMIT(locals,crit_params,crits,Some(attach_param,attach),
		   Some(detach_param,detach),attr)) ->
      Some(Ast.ADMIT(locals,crit_params,crits,
		     Some(attach_param,process_one_stm attach),
		     Some(detach_param,process_one_stm detach),attr))
  | Some(Ast.ADMIT(locals,crit_params,crits,None,None,attr)) as x -> x
  | _ -> raise (Error.Error "braces: unexpected admit")

let braces
    (Ast.SCHEDULER(nm,default,
		   cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		   trace,admit,handlers,ifunctions,functions,aspects,attr)) =
  Ast.SCHEDULER(nm,default,
		cstdefs,enums,procdecls,fundecls,valdefs,states,criteria,
		trace,admit,process_handlers handlers,
		process_functions ifunctions,
		  process_functions functions,process_aspects aspects,attr)
